
# minimax.jl: a simple example for numerical obtaining something like a minimax
# rational function approximation to a function on a given domain. Really just a
# toy example and should not be used in place of Remez.jl.
#

using StandaloneIpopt

function slack_constraint(tpq, n, x, fx)
  coef_p = @view tpq[2:(n+1)]
  coef_q = @view tpq[(n+2):end] # leading coef for q is one.
  err = evalpoly(x, coef_p)/evalpoly(x, coef_q) - fx
  err^2 - tpq[1]^2
end

@inline ols_init(x, fx, n) = [xk^j for xk in x, j in 0:(n-1)]\fx

function minimaxish(x::Vector{T}, fx::Vector{T}, n, m; kwargs...) where{T}
  # The constraints:
  N = length(x)
  cons = ntuple(j->(-1e22, 0.0)=>(tpq -> slack_constraint(tpq, n, x[j], fx[j])), N)
  # Setup using OLS-based inits and then optimize:
  init = vcat(5.0, ols_init(x, fx, n), vcat(1, zeros(m-1)))
  boxl = vcat(0.0,  fill(-1e22, n+m))
  res  = ipopt_optimize(tpq->tpq[1], init, cons; box_lower=boxl, kwargs...)
  stat = res.status
  in(stat, (0,1)) || @warn "Ipopt returned status $stat, you should investigate..."
  tpq = res.minimizer
  (max_error=tpq[1], p_coef=tpq[2:(n+1)], q_coef=tpq[(n+2):end])
end

function minimaxish(f, s, n, m; a=-1.0, b=1.0, kwargs...)
  cheb_nodes = [(b+a)/2 + ((b-a)/2)*cos(pi*(2*k-1)/(2*s)) for k in 1:s]
  fx = f.(cheb_nodes)
  minimaxish(cheb_nodes, fx, n, m; kwargs...)
end

# Sample use:
if !isinteractive()
  (max_err, p_coef, q_coef) = minimaxish(exp, 32, 8, 4; a=0.0, b=1.0, 
                                         max_iter=10_000, tol=1e-12)
  println("Maximum error of an (8,4) rational function approximation of exp \
          on (0.0, 1.0) with 32 Chebyshev nodes: $max_err")
  println("exp(0.33) - p(0.33)/q(0.33): \
          $(exp(0.33) - evalpoly(0.33, p_coef)/evalpoly(0.33, q_coef))")
end

