
function ipopt_optimize(obj, ini, constraints=noconstraints(); 
                        derivatives=AutoFwdDerivatives(obj, length(ini)),
                        conderivatives=AutoFwdConDerivatives(constraints),
                        box_lower=-1e22, box_upper=1e22, kwargs...)
  # bring in some default arguments:
  optargs = merge(DEF_IPOPT_KWARGS, Dict(kwargs))
  # get the dimension of the problem:
  arglen  = length(ini)
  # set up box constraints for the args and, if applicable, the constraints:
  box_l = prepare_bounds(arglen, box_lower)
  box_u = prepare_bounds(arglen, box_upper)
  # function for the constraints:
  constr_buf = zeros(constraints.ncon)
  # objective gradient and constraint jacobian:
  objg = (p,g)->_gradient(g, derivatives, p)
  jac_buf = zeros(constraints.ncon, arglen)
  jac_nz  = arglen*constraints.ncon
  constr_jac = (p,r,c,v) -> begin 
    isnothing(v) && return jac_structure!(r, c, constraints.ncon, length(p)) 
    _con_jacobian(jac_buf, constr_buf, conderivatives, p)
    v .= vec(permutedims(jac_buf))
  end
  # Lagrangian hessian:
  (lhessbuf, hesstmp) = (zeros(arglen, arglen), zeros(arglen, arglen))
  lagrange_h = (p,r,c,o,l,v) -> begin 
    isnothing(v) && return hes_structure!(r,c,length(p)) 
    if o != 0.0
      _hessian(lhessbuf, derivatives, p) 
    else
      fill!(lhessbuf, 0.0)
    end
    lv = view(l, 1:constraints.ncon)
    if constraints.ncon > 0 && !all(iszero, lv)
      _con_lagpart_hess(hesstmp, lv, conderivatives, p)
      lhessbuf .+= hesstmp
    end
    linear_buffer_ltri_colmajor!(v, lhessbuf) 
    nothing
  end
  # Create the final problem:
  constr_eval = (p, store) -> constraints.fun(store, p)
  prob = CreateIpoptProblem(arglen,  box_l, box_u,     # arg count and box
                            constraints.ncon, 
                            constraints.box_lower, 
                            constraints.box_upper,     # constr. count and box
                            jac_nz,                    # size of constr. jacobian
                            div(arglen*(arglen+1), 2), # size of hessian ltri
                            obj,                       # objective
                            constr_eval,               # constraint function
                            objg,                      # objective gradient
                            constr_jac,                # constraint jacobian
                            lagrange_h)                # lagrangian hessian
  # Add all the options:
  for (opt, val) in optargs
    ipopt_addoption!(prob, opt, val)
  end
  # Solve and return:
  prob.x = copy(ini)
  status = IpoptSolve(prob)
  (status=Ipopt._STATUS_CODES[status], minimizer=copy(prob.x), minval=obj(prob.x))
end

